@@ -27,6 +27,13 @@
 #include "libssh/crypto.h"
 #include "libssh/priv.h"
 #include "libssh/misc.h"
+#include "mbedcrypto-compat.h"
+#if defined(MBEDTLS_CHACHA20_C) && defined(MBEDTLS_POLY1305_C)
+#include "libssh/bytearray.h"
+#include "libssh/chacha20-poly1305-common.h"
+#include <mbedtls/chacha20.h>
+#include <mbedtls/poly1305.h>
+#endif
 
 #ifdef HAVE_LIBMBEDCRYPTO
 #include <mbedtls/md.h>
@@ -35,7 +42,7 @@
 #endif /* MBEDTLS_GCM_C */
 
 static mbedtls_entropy_context ssh_mbedtls_entropy;
-static mbedtls_ctr_drbg_context ssh_mbedtls_ctr_drbg;
+extern mbedtls_ctr_drbg_context ssh_mbedtls_ctr_drbg;
 
 static int libmbedcrypto_initialized = 0;
 
@@ -44,17 +51,40 @@ void ssh_reseed(void)
     mbedtls_ctr_drbg_reseed(&ssh_mbedtls_ctr_drbg, NULL, 0);
 }
 
-int ssh_get_random(void *where, int len, int strong)
+static mbedtls_md_type_t nid_to_md_algo(int nid)
 {
-    return ssh_mbedtls_random(where, len, strong);
+    switch (nid) {
+        case NID_mbedtls_nistp256:
+            return MBEDTLS_MD_SHA256;
+        case NID_mbedtls_nistp384:
+            return MBEDTLS_MD_SHA384;
+        case NID_mbedtls_nistp521:
+            return MBEDTLS_MD_SHA512;
+    }
+    return MBEDTLS_MD_NONE;
 }
 
-SHACTX sha1_init(void)
+void evp(int nid, unsigned char *digest, size_t len,
+        unsigned char *hash, unsigned int *hlen)
 {
-    SHACTX ctx = NULL;
-    int rc;
+    mbedtls_md_type_t algo = nid_to_md_algo(nid);
     const mbedtls_md_info_t *md_info =
-        mbedtls_md_info_from_type(MBEDTLS_MD_SHA1);
+        mbedtls_md_info_from_type(algo);
+
+
+    if (md_info != NULL) {
+        *hlen = mbedtls_md_get_size(md_info);
+        mbedtls_md(md_info, digest, len, hash);
+    }
+}
+
+EVPCTX evp_init(int nid)
+{
+    EVPCTX ctx = NULL;
+    int rc;
+    mbedtls_md_type_t algo = nid_to_md_algo(nid);
+    const mbedtls_md_info_t *md_info =
+        mbedtls_md_info_from_type(algo);
 
     if (md_info == NULL) {
         return NULL;
@@ -82,328 +112,17 @@ SHACTX sha1_init(void)
     return ctx;
 }
 
-void
-sha1_ctx_free(SHACTX c)
+void evp_update(EVPCTX ctx, const void *data, size_t len)
 {
-    mbedtls_md_free(c);
-    SAFE_FREE(c);
+    mbedtls_md_update(ctx, data, len);
 }
 
-int sha1_update(SHACTX c, const void *data, unsigned long len)
+void evp_final(EVPCTX ctx, unsigned char *md, unsigned int *mdlen)
 {
-    int rc = mbedtls_md_update(c, data, len);
-    if (rc != 0) {
-        return SSH_ERROR;
-    }
-    return SSH_OK;
-}
-
-int sha1_final(unsigned char *md, SHACTX c)
-{
-    int rc = mbedtls_md_finish(c, md);
-    sha1_ctx_free(c);
-    if (rc != 0) {
-        return SSH_ERROR;
-    }
-    return SSH_OK;
-}
-
-int sha1(const unsigned char *digest, int len, unsigned char *hash)
-{
-    const mbedtls_md_info_t *md_info =
-        mbedtls_md_info_from_type(MBEDTLS_MD_SHA1);
-    int rc;
-
-    if (md_info == NULL) {
-        return SSH_ERROR;
-    }
-    rc = mbedtls_md(md_info, digest, len, hash);
-    if (rc != 0) {
-        return SSH_ERROR;
-    }
-    return SSH_OK;
-}
-
-SHA256CTX sha256_init(void)
-{
-    SHA256CTX ctx = NULL;
-    int rc;
-    const mbedtls_md_info_t *md_info =
-        mbedtls_md_info_from_type(MBEDTLS_MD_SHA256);
-
-    if (md_info == NULL) {
-        return NULL;
-    }
-
-    ctx = malloc(sizeof(mbedtls_md_context_t));
-    if(ctx == NULL) {
-        return NULL;
-    }
-
-    mbedtls_md_init(ctx);
-
-    rc = mbedtls_md_setup(ctx, md_info, 0);
-    if (rc != 0) {
-        SAFE_FREE(ctx);
-        return NULL;
-    }
-
-    rc = mbedtls_md_starts(ctx);
-    if (rc != 0) {
-        SAFE_FREE(ctx);
-        return NULL;
-    }
-
-    return ctx;
-}
-
-void
-sha256_ctx_free(SHA256CTX c)
-{
-    mbedtls_md_free(c);
-    SAFE_FREE(c);
-}
-
-int sha256_update(SHA256CTX c, const void *data, unsigned long len)
-{
-    int rc = mbedtls_md_update(c, data, len);
-    if (rc != 0) {
-        return SSH_ERROR;
-    }
-    return SSH_OK;
-}
-
-int sha256_final(unsigned char *md, SHA256CTX c)
-{
-    int rc = mbedtls_md_finish(c, md);
-    sha256_ctx_free(c);
-    if (rc != 0) {
-        return SSH_ERROR;
-    }
-    return SSH_OK;
-}
-
-int sha256(const unsigned char *digest, int len, unsigned char *hash)
-{
-    const mbedtls_md_info_t *md_info =
-        mbedtls_md_info_from_type(MBEDTLS_MD_SHA256);
-    int rc;
-
-    if (md_info == NULL) {
-        return SSH_ERROR;
-    }
-    rc = mbedtls_md(md_info, digest, len, hash);
-    if (rc != 0) {
-        return SSH_ERROR;
-    }
-    return SSH_OK;
-}
-
-SHA384CTX sha384_init(void)
-{
-    SHA384CTX ctx = NULL;
-    int rc;
-    const mbedtls_md_info_t *md_info =
-        mbedtls_md_info_from_type(MBEDTLS_MD_SHA384);
-
-    if (md_info == NULL) {
-        return NULL;
-    }
-
-    ctx = malloc(sizeof(mbedtls_md_context_t));
-    if (ctx == NULL) {
-        return NULL;
-    }
-
-    mbedtls_md_init(ctx);
-
-    rc = mbedtls_md_setup(ctx, md_info, 0);
-    if (rc != 0) {
-        SAFE_FREE(ctx);
-        return NULL;
-    }
-
-    rc = mbedtls_md_starts(ctx);
-    if (rc != 0) {
-        SAFE_FREE(ctx);
-        return NULL;
-    }
-
-    return ctx;
-}
-
-void
-sha384_ctx_free(SHA384CTX c)
-{
-    mbedtls_md_free(c);
-    SAFE_FREE(c);
-}
-
-int sha384_update(SHA384CTX c, const void *data, unsigned long len)
-{
-    int rc = mbedtls_md_update(c, data, len);
-    if (rc != 0) {
-        return SSH_ERROR;
-    }
-    return SSH_OK;
-}
-
-int sha384_final(unsigned char *md, SHA384CTX c)
-{
-    int rc = mbedtls_md_finish(c, md);
-    sha384_ctx_free(c);
-    if (rc != 0) {
-        return SSH_ERROR;
-    }
-    return SSH_OK;
-}
-
-int sha384(const unsigned char *digest, int len, unsigned char *hash)
-{
-    const mbedtls_md_info_t *md_info =
-        mbedtls_md_info_from_type(MBEDTLS_MD_SHA384);
-    int rc;
-
-    if (md_info == NULL) {
-        return SSH_ERROR;
-    }
-    rc = mbedtls_md(md_info, digest, len, hash);
-    if (rc != 0) {
-        return SSH_ERROR;
-    }
-    return SSH_OK;
-}
-
-SHA512CTX sha512_init(void)
-{
-    SHA512CTX ctx = NULL;
-    int rc;
-    const mbedtls_md_info_t *md_info =
-        mbedtls_md_info_from_type(MBEDTLS_MD_SHA512);
-    if (md_info == NULL) {
-        return NULL;
-    }
-
-    ctx = malloc(sizeof(mbedtls_md_context_t));
-    if (ctx == NULL) {
-        return NULL;
-    }
-
-    mbedtls_md_init(ctx);
-
-    rc = mbedtls_md_setup(ctx, md_info, 0);
-    if (rc != 0) {
-        SAFE_FREE(ctx);
-        return NULL;
-    }
-
-    rc = mbedtls_md_starts(ctx);
-    if (rc != 0) {
-        SAFE_FREE(ctx);
-        return NULL;
-    }
-
-    return ctx;
-}
-
-void
-sha512_ctx_free(SHA512CTX c)
-{
-    mbedtls_md_free(c);
-    SAFE_FREE(c);
-}
-
-int sha512_update(SHA512CTX c, const void *data, unsigned long len)
-{
-    int rc = mbedtls_md_update(c, data, len);
-    if (rc != 0) {
-        return SSH_ERROR;
-    }
-    return SSH_OK;
-}
-
-int sha512_final(unsigned char *md, SHA512CTX c)
-{
-    int rc = mbedtls_md_finish(c, md);
-    sha512_ctx_free(c);
-    if (rc != 0) {
-        return SSH_ERROR;
-    }
-    return SSH_OK;
-}
-
-int sha512(const unsigned char *digest, int len, unsigned char *hash)
-{
-    const mbedtls_md_info_t *md_info =
-        mbedtls_md_info_from_type(MBEDTLS_MD_SHA512);
-    int rc;
-
-    if (md_info == NULL) {
-        return SSH_ERROR;
-    }
-    rc = mbedtls_md(md_info, digest, len, hash);
-    if (rc != 0) {
-        return SSH_ERROR;
-    }
-    return SSH_OK;
-}
-
-MD5CTX md5_init(void)
-{
-    MD5CTX ctx = NULL;
-    int rc;
-    const mbedtls_md_info_t *md_info =
-        mbedtls_md_info_from_type(MBEDTLS_MD_MD5);
-    if (md_info == NULL) {
-        return NULL;
-    }
-
-    ctx = malloc(sizeof(mbedtls_md_context_t));
-    if (ctx == NULL) {
-        return NULL;
-    }
-
-    mbedtls_md_init(ctx);
-
-    rc = mbedtls_md_setup(ctx, md_info, 0);
-    if (rc != 0) {
-        SAFE_FREE(ctx);
-        return NULL;
-    }
-
-    rc = mbedtls_md_starts(ctx);
-    if (rc != 0) {
-        SAFE_FREE(ctx);
-        return NULL;
-    }
-
-    return ctx;
-}
-
-void
-md5_ctx_free(MD5CTX c)
-{
-    mbedtls_md_free(c);
-    SAFE_FREE(c);
-}
-
-int md5_update(MD5CTX c, const void *data, unsigned long len)
-{
-    int rc = mbedtls_md_update(c, data, len);
-    if (rc != 0) {
-        return SSH_ERROR;
-    }
-    return SSH_OK;
-}
-
-int md5_final(unsigned char *md, MD5CTX c)
-{
-    int rc = mbedtls_md_finish(c, md);
-    md5_ctx_free(c);
-    if (rc != 0) {
-        return SSH_ERROR;
-    }
-    return SSH_OK;
+    *mdlen = mbedtls_md_get_size(ctx->MBEDTLS_PRIVATE(md_info));
+    mbedtls_md_finish(ctx, md);
+    mbedtls_md_free(ctx);
+    SAFE_FREE(ctx);
 }
 
 int ssh_kdf(struct ssh_crypto_struct *crypto,
@@ -415,7 +134,7 @@ int ssh_kdf(struct ssh_crypto_struct *crypto,
                              key_type, output, requested_len);
 }
 
-HMACCTX hmac_init(const void *key, int len, enum ssh_hmac_e type)
+HMACCTX hmac_init(const void *key, size_t len, enum ssh_hmac_e type)
 {
     HMACCTX ctx = NULL;
     const mbedtls_md_info_t *md_info = NULL;
@@ -464,17 +183,21 @@ error:
     return NULL;
 }
 
-void hmac_update(HMACCTX c, const void *data, unsigned long len)
+/* mbedtls returns 0 on success, but in this context
+ * success is 1 */
+int hmac_update(HMACCTX c, const void *data, size_t len)
 {
-    mbedtls_md_hmac_update(c, data, len);
+    return !mbedtls_md_hmac_update(c, data, len);
 }
 
-void hmac_final(HMACCTX c, unsigned char *hashmacbuf, unsigned int *len)
+int hmac_final(HMACCTX c, unsigned char *hashmacbuf, size_t *len)
 {
-    *len = mbedtls_md_get_size(c->md_info);
-    mbedtls_md_hmac_finish(c, hashmacbuf);
+    int rc;
+    *len = (unsigned int)mbedtls_md_get_size(c->md_info);
+    rc = !mbedtls_md_hmac_finish(c, hashmacbuf);
     mbedtls_md_free(c);
     SAFE_FREE(c);
+    return rc;
 }
 
 static int
@@ -485,6 +208,8 @@ cipher_init(struct ssh_cipher_struct *cipher,
 {
     const mbedtls_cipher_info_t *cipher_info = NULL;
     mbedtls_cipher_context_t *ctx;
+    size_t key_bitlen = 0;
+    size_t iv_size = 0;
     int rc;
 
     if (operation == MBEDTLS_ENCRYPT) {
@@ -505,15 +230,15 @@ cipher_init(struct ssh_cipher_struct *cipher,
         goto error;
     }
 
-    rc = mbedtls_cipher_setkey(ctx, key,
-                               cipher_info->key_bitlen,
-                               operation);
+    key_bitlen = mbedtls_cipher_info_get_key_bitlen(cipher_info);
+    rc = mbedtls_cipher_setkey(ctx, key, key_bitlen, operation);
     if (rc != 0) {
         SSH_LOG(SSH_LOG_WARNING, "mbedtls_cipher_setkey failed");
         goto error;
     }
 
-    rc = mbedtls_cipher_set_iv(ctx, IV, cipher_info->iv_size);
+    iv_size = mbedtls_cipher_info_get_iv_size(cipher_info);
+    rc = mbedtls_cipher_set_iv(ctx, IV, iv_size);
     if (rc != 0) {
         SSH_LOG(SSH_LOG_WARNING, "mbedtls_cipher_set_iv failed");
         goto error;
@@ -591,15 +316,16 @@ cipher_set_key_gcm(struct ssh_cipher_struct *cipher,
                    void *IV)
 {
     const mbedtls_cipher_info_t *cipher_info = NULL;
+    size_t key_bitlen = 0;
     int rc;
 
     mbedtls_gcm_init(&cipher->gcm_ctx);
     cipher_info = mbedtls_cipher_info_from_type(cipher->type);
 
-    rc = mbedtls_gcm_setkey(&cipher->gcm_ctx,
-                            MBEDTLS_CIPHER_ID_AES,
-                            key,
-                            cipher_info->key_bitlen);
+    key_bitlen = mbedtls_cipher_info_get_key_bitlen(cipher_info);
+    rc = mbedtls_gcm_setkey(&cipher->gcm_ctx, MBEDTLS_CIPHER_ID_AES,
+                            key, key_bitlen);
+
     if (rc != 0) {
         SSH_LOG(SSH_LOG_WARNING, "mbedtls_gcm_setkey failed");
         goto error;
@@ -905,6 +631,326 @@ cipher_decrypt_gcm(struct ssh_cipher_struct *cipher,
 }
 #endif /* MBEDTLS_GCM_C */
 
+#if defined(MBEDTLS_CHACHA20_C) && defined(MBEDTLS_POLY1305_C)
+
+struct chacha20_poly1305_keysched {
+    bool initialized;
+    /* cipher handle used for encrypting the packets */
+    mbedtls_chacha20_context main_ctx;
+    /* cipher handle used for encrypting the length field */
+    mbedtls_chacha20_context header_ctx;
+    /* Poly1305 key */
+    mbedtls_poly1305_context poly_ctx;
+};
+
+static void
+chacha20_poly1305_cleanup(struct ssh_cipher_struct *cipher)
+{
+    struct chacha20_poly1305_keysched *ctx = NULL;
+
+    if (cipher->chacha20_schedule == NULL) {
+        return;
+    }
+
+    ctx = cipher->chacha20_schedule;
+
+    if (ctx->initialized) {
+        mbedtls_chacha20_free(&ctx->main_ctx);
+        mbedtls_chacha20_free(&ctx->header_ctx);
+        mbedtls_poly1305_free(&ctx->poly_ctx);
+        ctx->initialized = false;
+    }
+
+    SAFE_FREE(cipher->chacha20_schedule);
+}
+
+static int
+chacha20_poly1305_set_key(struct ssh_cipher_struct *cipher,
+                          void *key,
+                          UNUSED_PARAM(void *IV))
+{
+    struct chacha20_poly1305_keysched *ctx = NULL;
+    uint8_t *u8key = key;
+    int ret = SSH_ERROR, rv;
+
+    if (cipher->chacha20_schedule == NULL) {
+        ctx = calloc(1, sizeof(*ctx));
+        if (ctx == NULL) {
+            return -1;
+        }
+        cipher->chacha20_schedule = ctx;
+    } else {
+        ctx = cipher->chacha20_schedule;
+    }
+
+    if (!ctx->initialized) {
+        mbedtls_chacha20_init(&ctx->main_ctx);
+        mbedtls_chacha20_init(&ctx->header_ctx);
+        mbedtls_poly1305_init(&ctx->poly_ctx);
+        ctx->initialized = true;
+    }
+
+    /* ChaCha20 keys initialization */
+    /* K2 uses the first half of the key */
+    rv = mbedtls_chacha20_setkey(&ctx->main_ctx, u8key);
+    if (rv != 0) {
+        SSH_LOG(SSH_LOG_WARNING, "mbedtls_chacha20_setkey(main_ctx) failed");
+        goto out;
+    }
+
+    /* K1 uses the second half of the key */
+    rv = mbedtls_chacha20_setkey(&ctx->header_ctx, u8key + CHACHA20_KEYLEN);
+    if (rv != 0) {
+        SSH_LOG(SSH_LOG_WARNING, "mbedtls_chacha20_setkey(header_ctx) failed");
+        goto out;
+    }
+
+    ret = SSH_OK;
+out:
+    if (ret != SSH_OK) {
+        chacha20_poly1305_cleanup(cipher);
+    }
+    return ret;
+}
+
+static const uint8_t zero_block[CHACHA20_BLOCKSIZE] = {0};
+
+static int
+chacha20_poly1305_set_iv(struct ssh_cipher_struct *cipher,
+                         uint64_t seq)
+{
+    struct chacha20_poly1305_keysched *ctx = cipher->chacha20_schedule;
+    uint8_t seqbuf[12] = {0};
+    int ret;
+
+    /* The nonce in mbedTLS is 96 b long. The counter is passed through separate
+     * parameter of 32 b size.
+     * Encode the seqence number into the last 8 bytes.
+     */
+    PUSH_BE_U64(seqbuf, 4, seq);
+#ifdef DEBUG_CRYPTO
+    ssh_log_hexdump("seqbuf (chacha20 IV)", seqbuf, sizeof(seqbuf));
+#endif /* DEBUG_CRYPTO */
+
+    ret = mbedtls_chacha20_starts(&ctx->header_ctx, seqbuf, 0);
+    if (ret != 0) {
+        SSH_LOG(SSH_LOG_WARNING, "mbedtls_chacha20_starts(header_ctx) failed");
+        return SSH_ERROR;
+    }
+
+    ret = mbedtls_chacha20_starts(&ctx->main_ctx, seqbuf, 0);
+    if (ret != 0) {
+        SSH_LOG(SSH_LOG_WARNING, "mbedtls_chacha20_starts(main_ctx) failed");
+        return SSH_ERROR;
+    }
+
+    return SSH_OK;
+}
+
+static int
+chacha20_poly1305_packet_setup(struct ssh_cipher_struct *cipher,
+                               uint64_t seq,
+                               int do_encrypt)
+{
+    struct chacha20_poly1305_keysched *ctx = cipher->chacha20_schedule;
+    uint8_t poly_key[CHACHA20_BLOCKSIZE];
+    int ret = SSH_ERROR, rv;
+
+    /* The initialization for decrypt was already done with the length block */
+    if (do_encrypt) {
+        rv = chacha20_poly1305_set_iv(cipher, seq);
+        if (rv != SSH_OK) {
+            return SSH_ERROR;
+        }
+    }
+
+    /* Output full ChaCha block so that counter increases by one for
+     * next step. */
+    rv = mbedtls_chacha20_update(&ctx->main_ctx, sizeof(zero_block),
+                                 zero_block, poly_key);
+    if (rv != 0) {
+        SSH_LOG(SSH_LOG_WARNING, "mbedtls_chacha20_update failed");
+        goto out;
+    }
+#ifdef DEBUG_CRYPTO
+    ssh_log_hexdump("poly_key", poly_key, POLY1305_KEYLEN);
+#endif /* DEBUG_CRYPTO */
+
+    /* Set the Poly1305 key */
+    rv = mbedtls_poly1305_starts(&ctx->poly_ctx, poly_key);
+    if (rv != 0) {
+        SSH_LOG(SSH_LOG_WARNING, "mbedtls_poly1305_starts failed");
+        goto out;
+    }
+
+    ret = SSH_OK;
+out:
+    explicit_bzero(poly_key, sizeof(poly_key));
+    return ret;
+}
+
+static int
+chacha20_poly1305_aead_decrypt_length(struct ssh_cipher_struct *cipher,
+                                      void *in,
+                                      uint8_t *out,
+                                      size_t len,
+                                      uint64_t seq)
+{
+    struct chacha20_poly1305_keysched *ctx = cipher->chacha20_schedule;
+    int rv;
+
+    if (len < sizeof(uint32_t)) {
+        return SSH_ERROR;
+    }
+
+#ifdef DEBUG_CRYPTO
+    ssh_log_hexdump("encrypted length", (uint8_t *)in, sizeof(uint32_t));
+#endif /* DEBUG_CRYPTO */
+
+    /* Set IV for the header context */
+    rv = chacha20_poly1305_set_iv(cipher, seq);
+    if (rv != SSH_OK) {
+        return SSH_ERROR;
+    }
+
+    rv = mbedtls_chacha20_update(&ctx->header_ctx, sizeof(uint32_t), in, out);
+    if (rv != 0) {
+        SSH_LOG(SSH_LOG_WARNING, "mbedtls_chacha20_update failed");
+        return SSH_ERROR;
+    }
+
+#ifdef DEBUG_CRYPTO
+    ssh_log_hexdump("deciphered length", out, sizeof(uint32_t));
+#endif /* DEBUG_CRYPTO */
+
+    return SSH_OK;
+}
+
+static int
+chacha20_poly1305_aead_decrypt(struct ssh_cipher_struct *cipher,
+                               void *complete_packet,
+                               uint8_t *out,
+                               size_t encrypted_size,
+                               uint64_t seq)
+{
+    struct chacha20_poly1305_keysched *ctx = cipher->chacha20_schedule;
+    uint8_t *mac = (uint8_t *)complete_packet + sizeof(uint32_t) +
+                   encrypted_size;
+    uint8_t tag[POLY1305_TAGLEN] = {0};
+    int ret = SSH_ERROR;
+    int rv, cmp = 0;
+
+    /* Prepare the Poly1305 key */
+    rv = chacha20_poly1305_packet_setup(cipher, seq, 0);
+    if (rv != SSH_OK) {
+        SSH_LOG(SSH_LOG_WARNING, "Failed to setup packet");
+        goto out;
+    }
+
+#ifdef DEBUG_CRYPTO
+    ssh_log_hexdump("received mac", mac, POLY1305_TAGLEN);
+#endif /* DEBUG_CRYPTO */
+
+    /* Calculate MAC of received data */
+    rv = mbedtls_poly1305_update(&ctx->poly_ctx, complete_packet,
+                                 encrypted_size + sizeof(uint32_t));
+    if (rv != 0) {
+        SSH_LOG(SSH_LOG_WARNING, "mbedtls_poly1305_update failed");
+        goto out;
+    }
+
+    rv = mbedtls_poly1305_finish(&ctx->poly_ctx, tag);
+    if (rv != 0) {
+        SSH_LOG(SSH_LOG_WARNING, "mbedtls_poly1305_finish failed");
+        goto out;
+    }
+
+#ifdef DEBUG_CRYPTO
+    ssh_log_hexdump("calculated mac", tag, POLY1305_TAGLEN);
+#endif /* DEBUG_CRYPTO */
+
+    /* Verify the calculated MAC matches the attached MAC */
+    cmp = secure_memcmp(tag, mac, POLY1305_TAGLEN);
+    if (cmp != 0) {
+        /* mac error */
+        SSH_LOG(SSH_LOG_PACKET, "poly1305 verify error");
+        return SSH_ERROR;
+    }
+
+    /* Decrypt the message */
+    rv = mbedtls_chacha20_update(&ctx->main_ctx, encrypted_size,
+                                 (uint8_t *)complete_packet + sizeof(uint32_t),
+                                 out);
+    if (rv != 0) {
+        SSH_LOG(SSH_LOG_WARNING, "mbedtls_chacha20_update failed");
+        goto out;
+    }
+
+    ret = SSH_OK;
+out:
+    return ret;
+}
+
+static void
+chacha20_poly1305_aead_encrypt(struct ssh_cipher_struct *cipher,
+                               void *in,
+                               void *out,
+                               size_t len,
+                               uint8_t *tag,
+                               uint64_t seq)
+{
+    struct ssh_packet_header *in_packet = in, *out_packet = out;
+    struct chacha20_poly1305_keysched *ctx = cipher->chacha20_schedule;
+    int ret;
+
+    /* Prepare the Poly1305 key */
+    ret = chacha20_poly1305_packet_setup(cipher, seq, 1);
+    if (ret != SSH_OK) {
+        SSH_LOG(SSH_LOG_WARNING, "Failed to setup packet");
+        return;
+    }
+
+#ifdef DEBUG_CRYPTO
+    ssh_log_hexdump("plaintext length",
+                    (unsigned char *)&in_packet->length, sizeof(uint32_t));
+#endif /* DEBUG_CRYPTO */
+    /* step 2, encrypt length field */
+    ret = mbedtls_chacha20_update(&ctx->header_ctx, sizeof(uint32_t),
+                                  (unsigned char *)&in_packet->length,
+                                  (unsigned char *)&out_packet->length);
+    if (ret != 0) {
+        SSH_LOG(SSH_LOG_WARNING, "mbedtls_chacha20_update failed");
+        return;
+    }
+#ifdef DEBUG_CRYPTO
+    ssh_log_hexdump("encrypted length",
+                    (unsigned char *)&out_packet->length, sizeof(uint32_t));
+#endif /* DEBUG_CRYPTO */
+
+    /* step 3, encrypt packet payload (main_ctx counter == 1) */
+    /* We already did encrypt one block so the counter should be in the correct position */
+    ret = mbedtls_chacha20_update(&ctx->main_ctx, len - sizeof(uint32_t),
+                                  in_packet->payload, out_packet->payload);
+    if (ret != 0) {
+        SSH_LOG(SSH_LOG_WARNING, "mbedtls_chacha20_update failed");
+        return;
+    }
+
+    /* step 4, compute the MAC */
+    ret = mbedtls_poly1305_update(&ctx->poly_ctx, (const unsigned char *)out_packet, len);
+    if (ret != 0) {
+        SSH_LOG(SSH_LOG_WARNING, "mbedtls_poly1305_update failed");
+        return;
+    }
+    ret = mbedtls_poly1305_finish(&ctx->poly_ctx, tag);
+    if (ret != 0) {
+        SSH_LOG(SSH_LOG_WARNING, "mbedtls_poly1305_finish failed");
+        return;
+    }
+}
+#endif /* defined(MBEDTLS_CHACHA20_C) && defined(MBEDTLS_POLY1305_C) */
+
+
 static void cipher_cleanup(struct ssh_cipher_struct *cipher)
 {
     mbedtls_cipher_free(&cipher->encrypt_ctx);
@@ -914,6 +960,17 @@ static void cipher_cleanup(struct ssh_cipher_struct *cipher)
 #endif /* MBEDTLS_GCM_C */
 }
 
+#ifdef WITH_INSECURE_NONE
+static void
+none_crypt(UNUSED_PARAM(struct ssh_cipher_struct *cipher),
+           void *in,
+           void *out,
+           size_t len)
+{
+    memcpy(out, in, len);
+}
+#endif /* WITH_INSECURE_NONE */
+
 static struct ssh_cipher_struct ssh_ciphertab[] = {
 #ifdef WITH_BLOWFISH_CIPHER
     {
@@ -1036,8 +1093,33 @@ static struct ssh_cipher_struct ssh_ciphertab[] = {
         .cleanup = cipher_cleanup
     },
     {
+#if defined(MBEDTLS_CHACHA20_C) && defined(MBEDTLS_POLY1305_C)
+        .ciphertype = SSH_AEAD_CHACHA20_POLY1305,
+        .name = "chacha20-poly1305@openssh.com",
+        .blocksize = 8,
+        .lenfield_blocksize = 4,
+        .keylen = sizeof(struct chacha20_poly1305_keysched),
+        .keysize = 2 * CHACHA20_KEYLEN * 8,
+        .tag_size = POLY1305_TAGLEN,
+        .set_encrypt_key = chacha20_poly1305_set_key,
+        .set_decrypt_key = chacha20_poly1305_set_key,
+        .aead_encrypt = chacha20_poly1305_aead_encrypt,
+        .aead_decrypt_length = chacha20_poly1305_aead_decrypt_length,
+        .aead_decrypt = chacha20_poly1305_aead_decrypt,
+        .cleanup = chacha20_poly1305_cleanup
+#else
         .name = "chacha20-poly1305@openssh.com"
+#endif
     },
+#ifdef WITH_INSECURE_NONE
+    {
+        .name = "none",
+        .blocksize = 8,
+        .keysize = 0,
+        .encrypt = none_crypt,
+        .decrypt = none_crypt,
+    },
+#endif /* WITH_INSECURE_NONE */
     {
         .name = NULL,
         .blocksize = 0,
@@ -1057,7 +1139,7 @@ struct ssh_cipher_struct *ssh_get_ciphertab(void)
 
 int ssh_crypto_init(void)
 {
-    size_t i;
+    UNUSED_VAR(size_t i);
     int rc;
 
     if (libmbedcrypto_initialized) {
@@ -1073,6 +1155,7 @@ int ssh_crypto_init(void)
         mbedtls_ctr_drbg_free(&ssh_mbedtls_ctr_drbg);
     }
 
+#if !(defined(MBEDTLS_CHACHA20_C) && defined(MBEDTLS_POLY1305_C))
     for (i = 0; ssh_ciphertab[i].name != NULL; i++) {
         int cmp;
 
@@ -1084,28 +1167,13 @@ int ssh_crypto_init(void)
             break;
         }
     }
+#endif
 
     libmbedcrypto_initialized = 1;
 
     return SSH_OK;
 }
 
-int ssh_mbedtls_random(void *where, int len, int strong)
-{
-    int rc = 0;
-    if (strong) {
-        mbedtls_ctr_drbg_set_prediction_resistance(&ssh_mbedtls_ctr_drbg,
-                MBEDTLS_CTR_DRBG_PR_ON);
-        rc = mbedtls_ctr_drbg_random(&ssh_mbedtls_ctr_drbg, where, len);
-        mbedtls_ctr_drbg_set_prediction_resistance(&ssh_mbedtls_ctr_drbg,
-                MBEDTLS_CTR_DRBG_PR_OFF);
-    } else {
-        rc = mbedtls_ctr_drbg_random(&ssh_mbedtls_ctr_drbg, where, len);
-    }
-
-    return !rc;
-}
-
 mbedtls_ctr_drbg_context *ssh_get_mbedtls_ctr_drbg_context(void)
 {
     return &ssh_mbedtls_ctr_drbg;
