@@ -32,7 +32,6 @@
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
-#include <net/if.h>
 
 #endif /* _WIN32 */
 
@@ -60,7 +59,6 @@
 #include <ws2tcpip.h>
 #include <shlobj.h>
 #include <direct.h>
-#include <netioapi.h>
 
 #ifdef HAVE_IO_H
 #include <io.h>
@@ -96,8 +94,6 @@
 #define ZLIB_STRING ""
 #endif
 
-#define ARPA_DOMAIN_MAX_LEN 63
-
 /**
  * @defgroup libssh_misc The SSH helper functions.
  * @ingroup libssh
@@ -108,8 +104,9 @@
  */
 
 #ifdef _WIN32
-char *ssh_get_user_home_dir(void) {
-  char tmp[MAX_PATH] = {0};
+char *ssh_get_user_home_dir(void)
+{
+  char tmp[PATH_MAX] = {0};
   char *szPath = NULL;
 
   if (SHGetSpecialFolderPathA(NULL, tmp, CSIDL_PROFILE, TRUE)) {
@@ -126,12 +123,13 @@ char *ssh_get_user_home_dir(void) {
 }
 
 /* we have read access on file */
-int ssh_file_readaccess_ok(const char *file) {
-  if (_access(file, 4) < 0) {
-    return 0;
-  }
+int ssh_file_readaccess_ok(const char *file)
+{
+    if (_access(file, 4) < 0) {
+        return 0;
+    }
 
-  return 1;
+    return 1;
 }
 
 /**
@@ -162,7 +160,8 @@ int ssh_dir_writeable(const char *path)
 #define SSH_USEC_IN_SEC         1000000LL
 #define SSH_SECONDS_SINCE_1601  11644473600LL
 
-int gettimeofday(struct timeval *__p, void *__t) {
+int ssh_gettimeofday(struct timeval *__p, void *__t)
+{
   union {
     unsigned long long ns100; /* time since 1 Jan 1601 in 100ns units */
     FILETIME ft;
@@ -175,7 +174,8 @@ int gettimeofday(struct timeval *__p, void *__t) {
   return (0);
 }
 
-char *ssh_get_local_username(void) {
+char *ssh_get_local_username(void)
+{
     DWORD size = 0;
     char *user;
 
@@ -194,7 +194,8 @@ char *ssh_get_local_username(void) {
     return NULL;
 }
 
-int ssh_is_ipaddr_v4(const char *str) {
+int ssh_is_ipaddr_v4(const char *str)
+{
     struct sockaddr_storage ss;
     int sslen = sizeof(ss);
     int rc = SOCKET_ERROR;
@@ -216,39 +217,25 @@ int ssh_is_ipaddr_v4(const char *str) {
     return 0;
 }
 
-int ssh_is_ipaddr(const char *str) {
+int ssh_is_ipaddr(const char *str)
+{
     int rc = SOCKET_ERROR;
-    char *s = strdup(str);
 
-    if (s == NULL) {
-        return -1;
-    }
-    if (strchr(s, ':')) {
+    if (strchr(str, ':')) {
         struct sockaddr_storage ss;
         int sslen = sizeof(ss);
-        char *network_interface = strchr(s, '%');
 
-        /* link-local (IP:v6:addr%ifname). */
-        if (network_interface != NULL) {
-            rc = if_nametoindex(network_interface + 1);
-            if (rc == 0) {
-                free(s);
-                return 0;
-            }
-            *network_interface = '\0';
-        }
-        rc = WSAStringToAddressA((LPSTR) s,
+        /* TODO link-local (IP:v6:addr%ifname). */
+        rc = WSAStringToAddressA((LPSTR) str,
                                  AF_INET6,
                                  NULL,
                                  (struct sockaddr*)&ss,
                                  &sslen);
         if (rc == 0) {
-            free(s);
             return 1;
         }
     }
 
-    free(s);
     return ssh_is_ipaddr_v4(str);
 }
 #else /* _WIN32 */
@@ -338,7 +325,8 @@ char *ssh_get_local_username(void)
     return name;
 }
 
-int ssh_is_ipaddr_v4(const char *str) {
+int ssh_is_ipaddr_v4(const char *str)
+{
     int rc = -1;
     struct in_addr dest;
 
@@ -350,40 +338,27 @@ int ssh_is_ipaddr_v4(const char *str) {
     return 0;
 }
 
-int ssh_is_ipaddr(const char *str) {
+int ssh_is_ipaddr(const char *str)
+{
     int rc = -1;
-    char *s = strdup(str);
 
-    if (s == NULL) {
-        return -1;
-    }
-    if (strchr(s, ':')) {
+    if (strchr(str, ':')) {
         struct in6_addr dest6;
-        char *network_interface = strchr(s, '%');
 
-        /* link-local (IP:v6:addr%ifname). */
-        if (network_interface != NULL) {
-            rc = if_nametoindex(network_interface + 1);
-            if (rc == 0) {
-                free(s);
-                return 0;
-            }
-            *network_interface = '\0';
-        }
-        rc = inet_pton(AF_INET6, s, &dest6);
+        /* TODO link-local (IP:v6:addr%ifname). */
+        rc = inet_pton(AF_INET6, str, &dest6);
         if (rc > 0) {
-            free(s);
             return 1;
         }
     }
 
-    free(s);
     return ssh_is_ipaddr_v4(str);
 }
 
 #endif /* _WIN32 */
 
-char *ssh_lowercase(const char* str) {
+char *ssh_lowercase(const char* str)
+{
   char *new, *p;
 
   if (str == NULL) {
@@ -426,15 +401,17 @@ char *ssh_hostport(const char *host, int port)
  * @brief Convert a buffer into a colon separated hex string.
  * The caller has to free the memory.
  *
- * @param  what         What should be converted to a hex string.
+ * @param[in]  what         What should be converted to a hex string.
  *
- * @param  len          Length of the buffer to convert.
+ * @param[in]  len          Length of the buffer to convert.
  *
- * @return              The hex string or NULL on error.
+ * @return                  The hex string or NULL on error. The memory needs
+ *                          to be freed using ssh_string_free_char().
  *
  * @see ssh_string_free_char()
  */
-char *ssh_get_hexa(const unsigned char *what, size_t len) {
+char *ssh_get_hexa(const unsigned char *what, size_t len)
+{
     const char h[] = "0123456789abcdef";
     char *hexa;
     size_t i;
@@ -462,7 +439,8 @@ char *ssh_get_hexa(const unsigned char *what, size_t len) {
 /**
  * @deprecated          Please use ssh_print_hash() instead
  */
-void ssh_print_hexa(const char *descr, const unsigned char *what, size_t len) {
+void ssh_print_hexa(const char *descr, const unsigned char *what, size_t len)
+{
     char *hexa = ssh_get_hexa(what, len);
 
     if (hexa == NULL) {
@@ -489,7 +467,7 @@ void ssh_print_hexa(const char *descr, const unsigned char *what, size_t len) {
  * "  00000000  00 01 02 03 04 05 06 07  08 09 0a 0b 0c 0d 0e 0f  ................"
  *
  * The value for each byte as corresponding ASCII character is printed at the
- * end if the value is printable. Otherwise it is replace with '.'.
+ * end if the value is printable. Otherwise, it is replaced with '.'.
  *
  * @param[in] descr A description for the content to be logged
  * @param[in] what  The buffer to be logged
@@ -683,48 +661,54 @@ error:
  *  }
  * @endcode
  */
-const char *ssh_version(int req_version) {
-  if (req_version <= LIBSSH_VERSION_INT) {
-    return SSH_STRINGIFY(LIBSSH_VERSION) GCRYPT_STRING CRYPTO_STRING MBED_STRING
-      ZLIB_STRING;
-  }
+const char *ssh_version(int req_version)
+{
+    if (req_version <= LIBSSH_VERSION_INT) {
+        return SSH_STRINGIFY(LIBSSH_VERSION) GCRYPT_STRING CRYPTO_STRING
+               MBED_STRING ZLIB_STRING;
+    }
 
-  return NULL;
-}
-
-struct ssh_list *ssh_list_new(void) {
-  struct ssh_list *ret=malloc(sizeof(struct ssh_list));
-  if(!ret)
     return NULL;
-  ret->root=ret->end=NULL;
-  return ret;
 }
 
-void ssh_list_free(struct ssh_list *list){
-  struct ssh_iterator *ptr,*next;
-  if(!list)
-    return;
-  ptr=list->root;
-  while(ptr){
-    next=ptr->next;
-    SAFE_FREE(ptr);
-    ptr=next;
-  }
-  SAFE_FREE(list);
+struct ssh_list *ssh_list_new(void)
+{
+    struct ssh_list *ret = malloc(sizeof(struct ssh_list));
+    if (!ret)
+        return NULL;
+    ret->root = ret->end = NULL;
+    return ret;
 }
 
-struct ssh_iterator *ssh_list_get_iterator(const struct ssh_list *list){
-  if(!list)
+void ssh_list_free(struct ssh_list *list)
+{
+    struct ssh_iterator *ptr, *next;
+    if (!list)
+        return;
+    ptr = list->root;
+    while (ptr) {
+        next = ptr->next;
+        SAFE_FREE(ptr);
+        ptr = next;
+    }
+    SAFE_FREE(list);
+}
+
+struct ssh_iterator *ssh_list_get_iterator(const struct ssh_list *list)
+{
+    if (!list)
+        return NULL;
+    return list->root;
+}
+
+struct ssh_iterator *ssh_list_find(const struct ssh_list *list, void *value)
+{
+    struct ssh_iterator *it;
+
+    for (it = ssh_list_get_iterator(list); it != NULL ; it = it->next)
+        if (it->data == value)
+            return it;
     return NULL;
-  return list->root;
-}
-
-struct ssh_iterator *ssh_list_find(const struct ssh_list *list, void *value){
-  struct ssh_iterator *it;
-  for(it = ssh_list_get_iterator(list); it != NULL ;it=it->next)
-    if(it->data==value)
-      return it;
-  return NULL;
 }
 
 /**
@@ -737,7 +721,7 @@ struct ssh_iterator *ssh_list_find(const struct ssh_list *list, void *value){
 size_t ssh_list_count(const struct ssh_list *list)
 {
   struct ssh_iterator *it = NULL;
-  int count = 0;
+  size_t count = 0;
 
   for (it = ssh_list_get_iterator(list); it != NULL ; it = it->next) {
       count++;
@@ -746,16 +730,19 @@ size_t ssh_list_count(const struct ssh_list *list)
   return count;
 }
 
-static struct ssh_iterator *ssh_iterator_new(const void *data){
-  struct ssh_iterator *iterator=malloc(sizeof(struct ssh_iterator));
-  if(!iterator)
-    return NULL;
-  iterator->next=NULL;
-  iterator->data=data;
-  return iterator;
+static struct ssh_iterator *ssh_iterator_new(const void *data)
+{
+    struct ssh_iterator *iterator = malloc(sizeof(struct ssh_iterator));
+
+    if (!iterator)
+        return NULL;
+    iterator->next = NULL;
+    iterator->data = data;
+    return iterator;
 }
 
-int ssh_list_append(struct ssh_list *list,const void *data){
+int ssh_list_append(struct ssh_list *list,const void *data)
+{
   struct ssh_iterator *iterator = NULL;
 
   if (list == NULL) {
@@ -778,7 +765,8 @@ int ssh_list_append(struct ssh_list *list,const void *data){
   return SSH_OK;
 }
 
-int ssh_list_prepend(struct ssh_list *list, const void *data){
+int ssh_list_prepend(struct ssh_list *list, const void *data)
+{
   struct ssh_iterator *it = NULL;
 
   if (list == NULL) {
@@ -802,8 +790,9 @@ int ssh_list_prepend(struct ssh_list *list, const void *data){
   return SSH_OK;
 }
 
-void ssh_list_remove(struct ssh_list *list, struct ssh_iterator *iterator){
-  struct ssh_iterator *ptr,*prev;
+void ssh_list_remove(struct ssh_list *list, struct ssh_iterator *iterator)
+{
+  struct ssh_iterator *ptr, *prev;
 
   if (list == NULL) {
       return;
@@ -842,7 +831,8 @@ void ssh_list_remove(struct ssh_list *list, struct ssh_iterator *iterator){
  * @returns             A pointer to the element being stored in head, or NULL
  *                      if the list is empty.
  */
-const void *_ssh_list_pop_head(struct ssh_list *list){
+const void *_ssh_list_pop_head(struct ssh_list *list)
+{
   struct ssh_iterator *iterator = NULL;
   const void *data = NULL;
 
@@ -868,17 +858,21 @@ const void *_ssh_list_pop_head(struct ssh_list *list){
  * dirname breaks a null-terminated pathname string into a directory component.
  * In the usual case, ssh_dirname() returns the string up to, but not including,
  * the final '/'. Trailing '/' characters are  not  counted as part of the
- * pathname. The caller must free the memory.
+ * pathname. The caller must free the memory using ssh_string_free_char().
  *
  * @param[in]  path     The path to parse.
  *
  * @return              The dirname of path or NULL if we can't allocate memory.
  *                      If path does not contain a slash, c_dirname() returns
- *                      the string ".".  If path is the string "/", it returns
+ *                      the string ".".  If path is a string "/", it returns
  *                      the string "/". If path is NULL or an empty string,
- *                      "." is returned.
+ *                      "." is returned. The memory needs to be freed using
+ *                      ssh_string_free_char().
+ *
+ * @see ssh_string_free_char()
  */
-char *ssh_dirname (const char *path) {
+char *ssh_dirname (const char *path)
+{
   char *new = NULL;
   size_t len;
 
@@ -929,11 +923,15 @@ char *ssh_dirname (const char *path) {
  * @param[in]  path     The path to parse.
  *
  * @return              The filename of path or NULL if we can't allocate
- *                      memory. If path is a the string "/", basename returns
+ *                      memory. If path is the string "/", basename returns
  *                      the string "/". If path is NULL or an empty string,
- *                      "." is returned.
+ *                      "." is returned. The caller needs to free this memory
+ *                      ssh_string_free_char().
+ *
+ * @see ssh_string_free_char()
  */
-char *ssh_basename (const char *path) {
+char *ssh_basename (const char *path)
+{
   char *new = NULL;
   const char *s;
   size_t len;
@@ -1066,9 +1064,13 @@ int ssh_mkdirs(const char *pathname, mode_t mode)
  *
  * @param[in]  d        The directory to expand.
  *
- * @return              The expanded directory, NULL on error.
+ * @return              The expanded directory, NULL on error. The caller
+ *                      needs to free the memory using ssh_string_free_char().
+ *
+ * @see ssh_string_free_char()
  */
-char *ssh_path_expand_tilde(const char *d) {
+char *ssh_path_expand_tilde(const char *d)
+{
     char *h = NULL, *r;
     const char *p;
     size_t ld;
@@ -1135,12 +1137,17 @@ char *ssh_path_expand_tilde(const char *d) {
  *              %l local hostname
  *              %r remote username
  *              %p remote port
- * @returns Expanded string.
+ * @returns Expanded string. The caller needs to free the memory using
+ *          ssh_string_free_char().
+ *
+ * @see ssh_string_free_char()
  */
-char *ssh_path_expand_escape(ssh_session session, const char *s) {
-    char host[NI_MAXHOST];
-    char buf[MAX_BUF_SIZE];
-    char *r, *x = NULL;
+char *ssh_path_expand_escape(ssh_session session, const char *s)
+{
+    char host[NI_MAXHOST] = {0};
+    char *buf = NULL;
+    char *r = NULL;
+    char *x = NULL;
     const char *p;
     size_t i, l;
 
@@ -1156,6 +1163,13 @@ char *ssh_path_expand_escape(ssh_session session, const char *s) {
         return NULL;
     }
 
+    buf = malloc(MAX_BUF_SIZE);
+    if (buf == NULL) {
+        ssh_set_error_oom(session);
+        free(r);
+        return NULL;
+    }
+
     p = r;
     buf[0] = '\0';
 
@@ -1165,6 +1179,7 @@ char *ssh_path_expand_escape(ssh_session session, const char *s) {
             buf[i] = *p;
             i++;
             if (i >= MAX_BUF_SIZE) {
+                free(buf);
                 free(r);
                 return NULL;
             }
@@ -1181,7 +1196,15 @@ char *ssh_path_expand_escape(ssh_session session, const char *s) {
             case '%':
                 goto escape;
             case 'd':
-                x = strdup(session->opts.sshdir);
+                if (session->opts.sshdir) {
+                    x = strdup(session->opts.sshdir);
+                } else {
+                    ssh_set_error(session, SSH_FATAL,
+                            "Cannot expand sshdir");
+                    free(buf);
+                    free(r);
+                    return NULL;
+                }
                 break;
             case 'u':
                 x = ssh_get_local_username();
@@ -1192,13 +1215,29 @@ char *ssh_path_expand_escape(ssh_session session, const char *s) {
                 }
                 break;
             case 'h':
-                x = strdup(session->opts.host);
+                if (session->opts.host) {
+                    x = strdup(session->opts.host);
+                } else {
+                    ssh_set_error(session, SSH_FATAL,
+                            "Cannot expand host");
+                    free(buf);
+                    free(r);
+                    return NULL;
+                }
                 break;
             case 'r':
-                x = strdup(session->opts.username);
+                if (session->opts.username) {
+                    x = strdup(session->opts.username);
+                } else {
+                    ssh_set_error(session, SSH_FATAL,
+                            "Cannot expand username");
+                    free(buf);
+                    free(r);
+                    return NULL;
+                }
                 break;
             case 'p':
-                {
+                if (session->opts.port > 0) {
                   char tmp[6];
 
                   snprintf(tmp, sizeof(tmp), "%hu",
@@ -1210,12 +1249,14 @@ char *ssh_path_expand_escape(ssh_session session, const char *s) {
             default:
                 ssh_set_error(session, SSH_FATAL,
                         "Wrong escape sequence detected");
+                free(buf);
                 free(r);
                 return NULL;
         }
 
         if (x == NULL) {
             ssh_set_error_oom(session);
+            free(buf);
             free(r);
             return NULL;
         }
@@ -1224,19 +1265,26 @@ char *ssh_path_expand_escape(ssh_session session, const char *s) {
         if (i >= MAX_BUF_SIZE) {
             ssh_set_error(session, SSH_FATAL,
                     "String too long");
+            free(buf);
             free(x);
             free(r);
             return NULL;
         }
         l = strlen(buf);
-        strncpy(buf + l, x, sizeof(buf) - l - 1);
+        strncpy(buf + l, x, MAX_BUF_SIZE - l - 1);
         buf[i] = '\0';
         SAFE_FREE(x);
     }
 
     free(r);
-    return strdup(buf);
-#undef MAX_BUF_SIZE
+
+    /* strip the unused space by realloc */
+    x = realloc(buf, strlen(buf) + 1);
+    if (x == NULL) {
+        ssh_set_error_oom(session);
+        free(buf);
+    }
+    return x;
 }
 
 /**
@@ -1312,28 +1360,33 @@ int ssh_analyze_banner(ssh_session session, int server)
          * 012345678901234567890
          */
         if (strlen(openssh) > 9) {
+            errno = 0;
             major = strtoul(openssh + 8, &tmp, 10);
             if ((tmp == (openssh + 8)) ||
                 ((errno == ERANGE) && (major == ULONG_MAX)) ||
                 ((errno != 0) && (major == 0)) ||
                 ((major < 1) || (major > 100))) {
                 /* invalid major */
+                errno = 0;
                 goto done;
             }
 
+            errno = 0;
             minor = strtoul(openssh + 10, &tmp, 10);
             if ((tmp == (openssh + 10)) ||
                 ((errno == ERANGE) && (major == ULONG_MAX)) ||
                 ((errno != 0) && (major == 0)) ||
                 (minor > 100)) {
                 /* invalid minor */
+                errno = 0;
                 goto done;
             }
 
             session->openssh = SSH_VERSION_INT(((int) major), ((int) minor), 0);
 
             SSH_LOG(SSH_LOG_PROTOCOL,
-                    "We are talking to an OpenSSH client version: %lu.%lu (%x)",
+                    "We are talking to an OpenSSH %s version: %lu.%lu (%x)",
+                    server ? "client" : "server",
                     major, minor, session->openssh);
         }
     }
@@ -1354,7 +1407,8 @@ done:
  * @brief initializes a timestamp to the current time
  * @param[out] ts pointer to an allocated ssh_timestamp structure
  */
-void ssh_timestamp_init(struct ssh_timestamp *ts){
+void ssh_timestamp_init(struct ssh_timestamp *ts)
+{
 #ifdef HAVE_CLOCK_GETTIME
   struct timespec tp;
   clock_gettime(CLOCK, &tp);
@@ -1377,17 +1431,18 @@ void ssh_timestamp_init(struct ssh_timestamp *ts){
  * @returns difference in milliseconds
  */
 
-static int ssh_timestamp_difference(struct ssh_timestamp *old,
-    struct ssh_timestamp *new){
-  long seconds, usecs, msecs;
-  seconds = new->seconds - old->seconds;
-  usecs = new->useconds - old->useconds;
-  if (usecs < 0){
-    seconds--;
-    usecs += 1000000;
-  }
-  msecs = seconds * 1000 + usecs/1000;
-  return msecs;
+static int
+ssh_timestamp_difference(struct ssh_timestamp *old, struct ssh_timestamp *new)
+{
+    long seconds, usecs, msecs;
+    seconds = new->seconds - old->seconds;
+    usecs = new->useconds - old->useconds;
+    if (usecs < 0){
+        seconds--;
+        usecs += 1000000;
+    }
+    msecs = seconds * 1000 + usecs/1000;
+    return msecs;
 }
 
 /**
@@ -1398,14 +1453,20 @@ static int ssh_timestamp_difference(struct ssh_timestamp *old,
  * @param[in] usec number of microseconds
  * @returns milliseconds, or 10000 if user supplied values are equal to zero
  */
-int ssh_make_milliseconds(long sec, long usec) {
-	int res = usec ? (usec / 1000) : 0;
+int ssh_make_milliseconds(unsigned long sec, unsigned long usec)
+{
+	unsigned long res = usec ? (usec / 1000) : 0;
 	res += (sec * 1000);
 	if (res == 0) {
 		res = 10 * 1000; /* use a reasonable default value in case
 				* SSH_OPTIONS_TIMEOUT is not set in options. */
 	}
-	return res;
+
+    if (res > INT_MAX) {
+        return SSH_TIMEOUT_INFINITE;
+    } else {
+        return (int)res;
+    }
 }
 
 /**
@@ -1418,7 +1479,8 @@ int ssh_make_milliseconds(long sec, long usec) {
  * @returns 1 if timeout is elapsed
  *          0 otherwise
  */
-int ssh_timeout_elapsed(struct ssh_timestamp *ts, int timeout) {
+int ssh_timeout_elapsed(struct ssh_timestamp *ts, int timeout)
+{
     struct ssh_timestamp now;
 
     switch(timeout) {
@@ -1450,7 +1512,8 @@ int ssh_timeout_elapsed(struct ssh_timestamp *ts, int timeout) {
  *             timeout
  * @returns   remaining time in milliseconds, 0 if elapsed, -1 if never.
  */
-int ssh_timeout_update(struct ssh_timestamp *ts, int timeout){
+int ssh_timeout_update(struct ssh_timestamp *ts, int timeout)
+{
   struct ssh_timestamp now;
   int ms, ret;
   if (timeout <= 0) {
@@ -1643,13 +1706,13 @@ int ssh_quote_file_name(const char *file_name, char *buf, size_t buf_len)
                 *dst++ = '\\';
                 break;
             case SINGLE_QUOTE:
-                /* Close the current quoted string and replace '!' for unquoted
+                /* Close the currently quoted string and replace '!' for unquoted
                  * "\!" */
                 *dst++ = '\'';
                 *dst++ = '\\';
                 break;
             case DOUBLE_QUOTE:
-                /* Close current quoted string and replace  "!" for unquoted
+                /* Close currently quoted string and replace  "!" for unquoted
                  * "\!" */
                 *dst++ = '"';
                 *dst++ = '\\';
@@ -1769,69 +1832,138 @@ int ssh_newline_vis(const char *string, char *buf, size_t buf_len)
 }
 
 /**
- * @brief Checks syntax of a domain name
+ * @internal
  *
- * The check is made based on the RFC1035 section 2.3.1
- * Allowed characters are: hyphen, period, digits (0-9) and letters (a-zA-Z)
+ * @brief Replaces the last 6 characters of a string from 'X' to 6 random hexdigits.
  *
- * The label should be no longer than 63 characters
- * The label should start with a letter and end with a letter or number
- * The label in this implementation can start with a number to allow virtual
- * URLs to pass. Note that this will make IPv4 addresses to pass
- * this check too.
- *
- * @param hostname The domain name to be checked, has to be null terminated
- *
- * @return SSH_OK if the hostname passes syntax check
- *         SSH_ERROR otherwise or if hostname is NULL or empty string
+ * @param[in,out]  template   Any input string with last 6 characters as 'X'.
+ * @returns -1 as error when the last 6 characters of the input to be replaced are not 'X'
+ * 0 otherwise.
  */
-int ssh_check_hostname_syntax(const char *hostname)
+int ssh_tmpname(char *template)
 {
-    char *it = NULL, *s = NULL, *buf = NULL;
-    size_t it_len;
-    char c;
+    char *tmp = NULL;
+    size_t i = 0;
+    int rc = 0;
+    uint8_t random[6];
 
-    if (hostname == NULL || strlen(hostname) == 0) {
-        return SSH_ERROR;
+    if (template == NULL) {
+        goto err;
     }
 
-    /* strtok_r writes into the string, keep the input clean */
-    s = strdup(hostname);
-    if (s == NULL) {
-        return SSH_ERROR;
+    tmp = template + strlen(template) - 6;
+    if (tmp < template) {
+        goto err;
     }
 
-    it = strtok_r(s, ".", &buf);
-    /* if the token has 0 length */
-    if (it == NULL) {
-        free(s);
-        return SSH_ERROR;
-    }
-    do {
-        it_len = strlen(it);
-        if (it_len > ARPA_DOMAIN_MAX_LEN ||
-            /* the first char must be a letter, but some virtual urls start
-             * with a number */
-            isalnum(it[0]) == 0 ||
-            isalnum(it[it_len - 1]) == 0) {
-            free(s);
-            return SSH_ERROR;
+    for (i = 0; i < 6; i++) {
+        if (tmp[i] != 'X') {
+            SSH_LOG(SSH_LOG_WARNING,
+                    "Invalid input. Last six characters of the input must be \'X\'");
+            goto err;
         }
-        while (*it != '\0') {
-            c = *it;
-            /* the "." is allowed too, but tokenization removes it from the
-             * string */
-            if (isalnum(c) == 0 && c != '-') {
-                free(s);
-                return SSH_ERROR;
-            }
-            it++;
+    }
+
+    rc = ssh_get_random(random, 6, 0);
+    if (!rc) {
+        SSH_LOG(SSH_LOG_WARNING,
+                "Could not generate random data\n");
+        goto err;
+    }
+
+    for (i = 0; i < 6; i++) {
+        /* Limit the random[i] < 32 */
+        random[i] &= 0x1f;
+        /* For values from 0 to 9 use numbers, otherwise use letters */
+        tmp[i] = random[i] > 9 ? random[i] + 'a' - 10 : random[i] + '0';
+    }
+
+    return 0;
+
+err:
+    errno = EINVAL;
+    return -1;
+}
+
+/**
+ * @internal
+ *
+ * @brief Finds the first occurrence of a pattern in a string and replaces it.
+ *
+ * @param[in]  src          Source string containing the pattern to be replaced.
+ * @param[in]  pattern      Pattern to be replaced in the source string.
+ *                          Note: this function replaces the first occurrence of
+ *                          pattern only.
+ * @param[in]  replace      String to be replaced is stored in replace.
+ *
+ * @returns  src_replaced a pointer that points to the replaced string.
+ * NULL if allocation fails or if src is NULL. The returned memory needs to be
+ * freed using ssh_string_free_char().
+ *
+ * @see ssh_string_free_char()
+ */
+char *ssh_strreplace(const char *src, const char *pattern, const char *replace)
+{
+    char *p = NULL;
+    char *src_replaced = NULL;
+
+    if (src == NULL) {
+        return NULL;
+    }
+
+    if (pattern == NULL || replace == NULL) {
+        return strdup(src);
+    }
+
+    p = strstr(src, pattern);
+
+    if (p != NULL) {
+        size_t offset = p - src;
+        size_t pattern_len = strlen(pattern);
+        size_t replace_len = strlen(replace);
+        size_t len  = strlen(src);
+        size_t len_replaced = len + replace_len - pattern_len + 1;
+
+        src_replaced = (char *)malloc(len_replaced);
+
+        if (src_replaced == NULL) {
+            return NULL;
         }
-    } while ((it = strtok_r(NULL, ".", &buf)) != NULL);
 
-    free(s);
+        memset(src_replaced, 0, len_replaced);
+        memcpy(src_replaced, src, offset);
+        memcpy(src_replaced + offset, replace, replace_len);
+        memcpy(src_replaced + offset + replace_len, src + offset + pattern_len, len - offset - pattern_len);
+        return src_replaced; /* free in the caller */
+    } else {
+        return strdup(src);
+    }
+}
 
-    return SSH_OK;
+/**
+ * @internal
+ *
+ * @brief Processes errno into error string
+ *
+ * @param[in] err_num The errno value
+ * @param[out] buf Pointer to a place where the string could be saved
+ * @param[in] buflen The allocated size of buf
+ *
+ * @return error string
+ */
+char *ssh_strerror(int err_num, char *buf, size_t buflen)
+{
+#if defined(_WIN32)
+    strerror_s(buf, buflen, err_num);
+    return buf;
+#elif defined(__linux__) && defined(_GNU_SOURCE)
+    /* GNU extension on Linux */
+    return strerror_r(err_num, buf, buflen);
+#else
+    /* POSIX version available for example on FreeBSD */
+    strerror_r(err_num, buf, buflen);
+    return buf;
+#endif /* _WIN32 */
 }
 
 /** @} */
