@@ -283,6 +283,23 @@ static int passphrase_to_key(char *data, unsigned int datalen,
   return 0;
 }
 
+void pki_key_clean(ssh_key key)
+{
+    if (key == NULL)
+        return;
+
+    if (key->dsa)
+        gcry_sexp_release(key->dsa);
+    if (key->rsa)
+        gcry_sexp_release(key->rsa);
+    if (key->ecdsa)
+        gcry_sexp_release(key->ecdsa);
+
+    key->dsa = NULL;
+    key->rsa = NULL;
+    key->ecdsa = NULL;
+}
+
 static int privatekey_decrypt(int algo, int mode, unsigned int key_len,
                        unsigned char *iv, unsigned int iv_len,
                        ssh_buffer data, ssh_auth_callback cb,
@@ -1506,11 +1523,13 @@ int pki_key_compare(const ssh_key k1,
             }
             break;
         case SSH_KEYTYPE_ED25519:
+        case SSH_KEYTYPE_SK_ED25519:
 		/* ed25519 keys handled globaly */
 		return 0;
         case SSH_KEYTYPE_ECDSA_P256:
         case SSH_KEYTYPE_ECDSA_P384:
         case SSH_KEYTYPE_ECDSA_P521:
+        case SSH_KEYTYPE_SK_ECDSA:
 #ifdef HAVE_GCRYPT_ECC
             if (k1->ecdsa_nid != k2->ecdsa_nid) {
                 return 1;
@@ -1533,7 +1552,9 @@ int pki_key_compare(const ssh_key k1,
         case SSH_KEYTYPE_ECDSA_P256_CERT01:
         case SSH_KEYTYPE_ECDSA_P384_CERT01:
         case SSH_KEYTYPE_ECDSA_P521_CERT01:
+        case SSH_KEYTYPE_SK_ECDSA_CERT01:
         case SSH_KEYTYPE_ED25519_CERT01:
+        case SSH_KEYTYPE_SK_ED25519_CERT01:
         case SSH_KEYTYPE_RSA1:
         case SSH_KEYTYPE_UNKNOWN:
             return 1;
@@ -1675,14 +1696,20 @@ ssh_string pki_publickey_to_blob(const ssh_key key)
 
             break;
         case SSH_KEYTYPE_ED25519:
-		rc = pki_ed25519_public_key_to_blob(buffer, key);
-		if (rc != SSH_OK){
-			goto fail;
-		}
-		break;
+        case SSH_KEYTYPE_SK_ED25519:
+            rc = pki_ed25519_public_key_to_blob(buffer, key);
+            if (rc != SSH_OK){
+                goto fail;
+            }
+            if (key->type == SSH_KEYTYPE_SK_ED25519 &&
+                ssh_buffer_add_ssh_string(buffer, key->sk_application) < 0) {
+                goto fail;
+            }
+            break;
         case SSH_KEYTYPE_ECDSA_P256:
         case SSH_KEYTYPE_ECDSA_P384:
         case SSH_KEYTYPE_ECDSA_P521:
+        case SSH_KEYTYPE_SK_ECDSA:
 #ifdef HAVE_GCRYPT_ECC
             type_s = ssh_string_from_char(
                        pki_key_ecdsa_nid_to_char(key->ecdsa_nid));
@@ -1713,6 +1740,12 @@ ssh_string pki_publickey_to_blob(const ssh_key key)
             ssh_string_burn(e);
             SSH_STRING_FREE(e);
             e = NULL;
+
+            if (key->type == SSH_KEYTYPE_SK_ECDSA &&
+                ssh_buffer_add_ssh_string(buffer, key->sk_application) < 0) {
+                goto fail;
+            }
+
             break;
 #endif
         case SSH_KEYTYPE_RSA1:
@@ -1978,8 +2011,9 @@ ssh_signature pki_signature_from_blob(const ssh_key pubkey,
 
             if (len > rsalen) {
                 SSH_LOG(SSH_LOG_WARN,
-                        "Signature is to big size: %lu",
-                        (unsigned long)len);
+                        "Signature is too big: %lu > %lu",
+                        (unsigned long)len,
+                        (unsigned long)rsalen);
                 ssh_signature_free(sig);
                 return NULL;
             }
@@ -2007,6 +2041,7 @@ ssh_signature pki_signature_from_blob(const ssh_key pubkey,
             }
             break;
         case SSH_KEYTYPE_ED25519:
+        case SSH_KEYTYPE_SK_ED25519:
 		rc = pki_signature_from_ed25519_blob(sig, sig_blob);
 		if (rc != SSH_OK){
 			ssh_signature_free(sig);
@@ -2016,6 +2051,7 @@ ssh_signature pki_signature_from_blob(const ssh_key pubkey,
         case SSH_KEYTYPE_ECDSA_P256:
         case SSH_KEYTYPE_ECDSA_P384:
         case SSH_KEYTYPE_ECDSA_P521:
+        case SSH_KEYTYPE_SK_ECDSA:
 #ifdef HAVE_GCRYPT_ECC
             { /* build ecdsa siganature */
                 ssh_buffer b;
@@ -2358,7 +2394,9 @@ int pki_verify_data_signature(ssh_signature signature,
         break;
     case SSH_DIGEST_AUTO:
         if (pubkey->type == SSH_KEYTYPE_ED25519 ||
-            pubkey->type == SSH_KEYTYPE_ED25519_CERT01)
+            pubkey->type == SSH_KEYTYPE_ED25519_CERT01 ||
+            pubkey->type == SSH_KEYTYPE_SK_ED25519 ||
+            pubkey->type == SSH_KEYTYPE_SK_ED25519_CERT01)
         {
             verify_input = input;
             hlen = input_len;
@@ -2427,6 +2465,8 @@ int pki_verify_data_signature(ssh_signature signature,
         case SSH_KEYTYPE_ECDSA_P256_CERT01:
         case SSH_KEYTYPE_ECDSA_P384_CERT01:
         case SSH_KEYTYPE_ECDSA_P521_CERT01:
+        case SSH_KEYTYPE_SK_ECDSA:
+        case SSH_KEYTYPE_SK_ECDSA_CERT01:
 #ifdef HAVE_GCRYPT_ECC
             err = gcry_sexp_build(&sexp,
                                   NULL,
@@ -2454,6 +2494,8 @@ int pki_verify_data_signature(ssh_signature signature,
 #endif
         case SSH_KEYTYPE_ED25519:
         case SSH_KEYTYPE_ED25519_CERT01:
+        case SSH_KEYTYPE_SK_ED25519:
+        case SSH_KEYTYPE_SK_ED25519_CERT01:
             rc = pki_ed25519_verify(pubkey, signature, verify_input, hlen);
             if (rc != SSH_OK) {
                 SSH_LOG(SSH_LOG_TRACE, "ED25519 error: Signature invalid");
@@ -2470,4 +2512,44 @@ int pki_verify_data_signature(ssh_signature signature,
     return SSH_OK;
 }
 
+int ssh_key_size(ssh_key key)
+{
+    switch (key->type) {
+    case SSH_KEYTYPE_DSS:
+    case SSH_KEYTYPE_DSS_CERT01:
+        return gcry_pk_get_nbits(key->dsa);
+    case SSH_KEYTYPE_RSA:
+    case SSH_KEYTYPE_RSA_CERT01:
+    case SSH_KEYTYPE_RSA1:
+        return gcry_pk_get_nbits(key->rsa);
+    case SSH_KEYTYPE_ECDSA_P256:
+    case SSH_KEYTYPE_ECDSA_P256_CERT01:
+    case SSH_KEYTYPE_ECDSA_P384:
+    case SSH_KEYTYPE_ECDSA_P384_CERT01:
+    case SSH_KEYTYPE_ECDSA_P521:
+    case SSH_KEYTYPE_ECDSA_P521_CERT01:
+    case SSH_KEYTYPE_SK_ECDSA:
+    case SSH_KEYTYPE_SK_ECDSA_CERT01:
+        return gcry_pk_get_nbits(key->ecdsa);
+    case SSH_KEYTYPE_ED25519:
+    case SSH_KEYTYPE_ED25519_CERT01:
+    case SSH_KEYTYPE_SK_ED25519:
+    case SSH_KEYTYPE_SK_ED25519_CERT01:
+        /* ed25519 keys have fixed size */
+        return 255;
+    case SSH_KEYTYPE_UNKNOWN:
+    default:
+        return SSH_ERROR;
+    }
+}
+
+int pki_uri_import(const char *uri_name, ssh_key *key, enum ssh_key_e key_type)
+{
+    (void) uri_name;
+    (void) key;
+    (void) key_type;
+    SSH_LOG(SSH_LOG_WARN,
+            "gcrypt does not support PKCS #11");
+    return SSH_ERROR;
+}
 #endif /* HAVE_LIBGCRYPT */
