@@ -25,7 +25,7 @@
  * Why a wrapper?
  *
  * Let's say you want to port libssh from libcrypto of openssl to libfoo
- * you are going to spend hours to remove every references to SHA1_Update()
+ * you are going to spend hours removing every reference to SHA1_Update()
  * to libfoo_sha1_update after the work is finished, you're going to have
  * only this file to modify it's not needed to say that your modifications
  * are welcome.
@@ -66,6 +66,9 @@ static struct ssh_hmac_struct ssh_hmac_tab[] = {
   { "hmac-sha2-256-etm@openssh.com", SSH_HMAC_SHA256,        true  },
   { "hmac-sha2-512-etm@openssh.com", SSH_HMAC_SHA512,        true  },
   { "hmac-md5-etm@openssh.com",      SSH_HMAC_MD5,           true  },
+#ifdef WITH_INSECURE_NONE
+  { "none",                          SSH_HMAC_NONE,          false },
+#endif /* WITH_INSECURE_NONE */
   { NULL,                            0,                      false }
 };
 
@@ -105,7 +108,7 @@ const char *ssh_hmac_type_to_string(enum ssh_hmac_e hmac_type, bool etm)
 }
 
 /* it allocates a new cipher structure based on its offset into the global table */
-static struct ssh_cipher_struct *cipher_new(int offset) {
+static struct ssh_cipher_struct *cipher_new(uint8_t offset) {
   struct ssh_cipher_struct *cipher = NULL;
 
   cipher = malloc(sizeof(struct ssh_cipher_struct));
@@ -147,16 +150,15 @@ static void cipher_free(struct ssh_cipher_struct *cipher) {
   SAFE_FREE(cipher);
 }
 
-struct ssh_crypto_struct *crypto_new(void)
-{
-    struct ssh_crypto_struct *crypto;
+struct ssh_crypto_struct *crypto_new(void) {
+   struct ssh_crypto_struct *crypto;
 
-    crypto = malloc(sizeof(struct ssh_crypto_struct));
-    if (crypto == NULL) {
-        return NULL;
-    }
-    ZERO_STRUCTP(crypto);
-    return crypto;
+  crypto = malloc(sizeof(struct ssh_crypto_struct));
+  if (crypto == NULL) {
+    return NULL;
+  }
+  ZERO_STRUCTP(crypto);
+  return crypto;
 }
 
 void crypto_free(struct ssh_crypto_struct *crypto)
@@ -176,7 +178,15 @@ void crypto_free(struct ssh_crypto_struct *crypto)
     SAFE_FREE(crypto->ecdh_server_pubkey);
     if(crypto->ecdh_privkey != NULL){
 #ifdef HAVE_OPENSSL_ECC
+/* TODO Change to new API when the OpenSSL will support export of uncompressed EC keys
+ * https://github.com/openssl/openssl/pull/16624
+ * #if OPENSSL_VERSION_NUMBER < 0x30000000L
+ */
+#if 1
         EC_KEY_free(crypto->ecdh_privkey);
+#else
+        EVP_PKEY_free(crypto->ecdh_privkey);
+#endif /* OPENSSL_VERSION_NUMBER */
 #elif defined HAVE_GCRYPT_ECC
         gcry_sexp_release(crypto->ecdh_privkey);
 #endif
@@ -237,7 +247,7 @@ static int crypt_set_algorithms2(ssh_session session)
     const char *wanted = NULL;
     struct ssh_cipher_struct *ssh_ciphertab=ssh_get_ciphertab();
     struct ssh_hmac_struct *ssh_hmactab=ssh_get_hmactab();
-    size_t i = 0;
+    uint8_t i = 0;
     int cmp;
 
     /*
