@@ -38,8 +38,6 @@
 #include <fcntl.h>
 #include <pwd.h>
 
-#define KEX_RETRY 32
-
 static uint64_t bytes = 2048; /* 2KB (more than the authentication phase) */
 
 static int sshd_setup(void **state)
@@ -148,29 +146,6 @@ static void torture_rekey_default(void **state)
     ssh_disconnect(s->ssh.session);
 }
 
-static void sanity_check_session(void **state)
-{
-    struct torture_state *s = *state;
-    struct ssh_crypto_struct *c = NULL;
-
-    c = s->ssh.session->current_crypto;
-    assert_non_null(c);
-    assert_int_equal(c->in_cipher->max_blocks,
-                     bytes / c->in_cipher->blocksize);
-    assert_int_equal(c->out_cipher->max_blocks,
-                     bytes / c->out_cipher->blocksize);
-    /* when strict kex is used, the newkeys reset the sequence number */
-    if ((s->ssh.session->flags & SSH_SESSION_FLAG_KEX_STRICT) != 0) {
-        assert_int_equal(c->out_cipher->packets, s->ssh.session->send_seq);
-        assert_int_equal(c->in_cipher->packets, s->ssh.session->recv_seq);
-    } else {
-        /* Otherwise we have less encrypted packets than transferred
-         * (first are not encrypted) */
-        assert_true(c->out_cipher->packets < s->ssh.session->send_seq);
-        assert_true(c->in_cipher->packets < s->ssh.session->recv_seq);
-    }
-}
-
 /* We lower the rekey limits manually and check that the rekey
  * really happens when sending data
  */
@@ -189,10 +164,16 @@ static void torture_rekey_send(void **state)
     rc = ssh_connect(s->ssh.session);
     assert_ssh_return_code(s->ssh.session, rc);
 
-    sanity_check_session(state);
-    /* Copy the initial secret hash = session_id so we know we changed keys later */
+    /* The blocks limit is set correctly */
     c = s->ssh.session->current_crypto;
-    assert_non_null(c);
+    assert_int_equal(c->in_cipher->max_blocks,
+                     bytes / c->in_cipher->blocksize);
+    assert_int_equal(c->out_cipher->max_blocks,
+                     bytes / c->out_cipher->blocksize);
+    /* We should have less encrypted packets than transfered (first are not encrypted) */
+    assert_true(c->out_cipher->packets < s->ssh.session->send_seq);
+    assert_true(c->in_cipher->packets < s->ssh.session->recv_seq);
+    /* Copy the initial secret hash = session_id so we know we changed keys later */
     secret_hash = malloc(c->digest_len);
     assert_non_null(secret_hash);
     memcpy(secret_hash, c->secret_hash, c->digest_len);
@@ -209,11 +190,10 @@ static void torture_rekey_send(void **state)
     rc = ssh_userauth_publickey_auto(s->ssh.session, NULL, NULL);
     assert_int_equal(rc, SSH_AUTH_SUCCESS);
 
-    /* send ignore packets of up to 1KB to trigger rekey. Send little bit more
-     * to make sure it completes with all different ciphers */
+    /* send ignore packets of up to 1KB to trigger rekey */
     memset(data, 0, sizeof(data));
     memset(data, 'A', 128);
-    for (i = 0; i < KEX_RETRY; i++) {
+    for (i = 0; i < 16; i++) {
         ssh_send_ignore(s->ssh.session, data);
         ssh_handle_packets(s->ssh.session, 50);
     }
@@ -288,11 +268,16 @@ static void torture_rekey_recv(void **state)
     int fd;
     sftp_file file;
     mode_t mask;
+    int rc;
 
-    sanity_check_session(state);
-    /* Copy the initial secret hash = session_id so we know we changed keys later */
+    /* The blocks limit is set correctly */
     c = s->ssh.session->current_crypto;
-    assert_non_null(c);
+    assert_int_equal(c->in_cipher->max_blocks, bytes / c->in_cipher->blocksize);
+    assert_int_equal(c->out_cipher->max_blocks, bytes / c->out_cipher->blocksize);
+    /* We should have less encrypted packets than transfered (first are not encrypted) */
+    assert_true(c->out_cipher->packets < s->ssh.session->send_seq);
+    assert_true(c->in_cipher->packets < s->ssh.session->recv_seq);
+    /* Copy the initial secret hash = session_id so we know we changed keys later */
     secret_hash = malloc(c->digest_len);
     assert_non_null(secret_hash);
     memcpy(secret_hash, c->secret_hash, c->digest_len);
@@ -317,6 +302,8 @@ static void torture_rekey_recv(void **state)
         assert_int_equal(byteswritten, bytesread);
     }
 
+    rc = sftp_close(file);
+    assert_int_equal(rc, SSH_NO_ERROR);
     close(fd);
 
     /* The rekey limit was restored in the new crypto to the same value */
@@ -477,10 +464,15 @@ static void torture_rekey_different_kex(void **state)
     assert_ssh_return_code(s->ssh.session, rc);
 
     /* The blocks limit is set correctly */
-    sanity_check_session(state);
-    /* Copy the initial secret hash = session_id so we know we changed keys later */
     c = s->ssh.session->current_crypto;
-    assert_non_null(c);
+    assert_int_equal(c->in_cipher->max_blocks,
+                     bytes / c->in_cipher->blocksize);
+    assert_int_equal(c->out_cipher->max_blocks,
+                     bytes / c->out_cipher->blocksize);
+    /* We should have less encrypted packets than transfered (first are not encrypted) */
+    assert_true(c->out_cipher->packets < s->ssh.session->send_seq);
+    assert_true(c->in_cipher->packets < s->ssh.session->recv_seq);
+    /* Copy the initial secret hash = session_id so we know we changed keys later */
     secret_hash = malloc(c->digest_len);
     assert_non_null(secret_hash);
     memcpy(secret_hash, c->secret_hash, c->digest_len);
@@ -507,15 +499,9 @@ static void torture_rekey_different_kex(void **state)
      * to make sure the rekey it completes with all different ciphers (paddings */
     memset(data, 0, sizeof(data));
     memset(data, 'A', 128);
-    for (i = 0; i < KEX_RETRY; i++) {
+    for (i = 0; i < 20; i++) {
         ssh_send_ignore(s->ssh.session, data);
-        ssh_handle_packets(s->ssh.session, 100);
-
-        c = s->ssh.session->current_crypto;
-        /* SHA256 len */
-        if (c->digest_len != 32) {
-            break;
-        }
+        ssh_handle_packets(s->ssh.session, 50);
     }
 
     /* The rekey limit was restored in the new crypto to the same value */
@@ -585,15 +571,9 @@ static void torture_rekey_server_different_kex(void **state)
      * to make sure the rekey it completes with all different ciphers (paddings */
     memset(data, 0, sizeof(data));
     memset(data, 'A', 128);
-    for (i = 0; i < KEX_RETRY; i++) {
+    for (i = 0; i < 25; i++) {
         ssh_send_ignore(s->ssh.session, data);
-        ssh_handle_packets(s->ssh.session, 100);
-
-        c = s->ssh.session->current_crypto;
-        /* SHA256 len */
-        if (c->digest_len != 32) {
-            break;
-        }
+        ssh_handle_packets(s->ssh.session, 50);
     }
 
     /* Check that the secret hash is different than initially */
@@ -634,6 +614,7 @@ static void torture_rekey_server_recv(void **state)
     int fd;
     sftp_file file;
     mode_t mask;
+    int rc;
 
     /* Copy the initial secret hash = session_id so we know we changed keys later */
     c = s->ssh.session->current_crypto;
@@ -661,6 +642,8 @@ static void torture_rekey_server_recv(void **state)
         assert_int_equal(byteswritten, bytesread);
     }
 
+    rc = sftp_close(file);
+    assert_int_equal(rc, SSH_NO_ERROR);
     close(fd);
 
     /* Check that the secret hash is different than initially */
@@ -674,92 +657,6 @@ static void torture_rekey_server_recv(void **state)
 #endif /* WITH_SFTP */
 
 
-static void setup_server_for_good_guess(void *state)
-{
-    const char *default_sshd_config = "KexAlgorithms curve25519-sha256";
-    const char *fips_sshd_config = "KexAlgorithms ecdh-sha2-nistp256";
-    const char *sshd_config = default_sshd_config;
-
-    if (ssh_fips_mode()) {
-        sshd_config = fips_sshd_config;
-    }
-    /* This sets an only supported kex algorithm that we do not have as a first
-    * option */
-    torture_update_sshd_config(state, sshd_config);
-}
-
-static void torture_rekey_guess_send(void **state)
-{
-    struct torture_state *s = *state;
-
-    setup_server_for_good_guess(state);
-
-    /* Make the client send the first_kex_packet_follows flag during key
-     * exchange as well as during the rekey */
-    s->ssh.session->send_first_kex_follows = true;
-
-    torture_rekey_send(state);
-}
-
-static void torture_rekey_guess_wrong_send(void **state)
-{
-    struct torture_state *s = *state;
-    const char *sshd_config = "KexAlgorithms diffie-hellman-group14-sha256";
-
-    /* This sets an only supported kex algorithm that we do not have as a first
-     * option */
-    torture_update_sshd_config(state, sshd_config);
-
-    /* Make the client send the first_kex_packet_follows flag during key
-     * exchange as well as during the rekey */
-    s->ssh.session->send_first_kex_follows = true;
-
-    torture_rekey_send(state);
-}
-
-#ifdef WITH_SFTP
-static void torture_rekey_guess_recv(void **state)
-{
-    struct torture_state *s = *state;
-    int rc;
-
-    setup_server_for_good_guess(state);
-
-    /* Make the client send the first_kex_packet_follows flag during key
-     * exchange as well as during the rekey */
-    s->ssh.session->send_first_kex_follows = true;
-
-    rc = ssh_options_set(s->ssh.session, SSH_OPTIONS_REKEY_DATA, &bytes);
-    assert_ssh_return_code(s->ssh.session, rc);
-
-    session_setup_sftp(state);
-
-    torture_rekey_recv(state);
-}
-
-static void torture_rekey_guess_wrong_recv(void **state)
-{
-    struct torture_state *s = *state;
-    const char *sshd_config = "KexAlgorithms diffie-hellman-group14-sha256";
-    int rc;
-
-    /* This sets an only supported kex algorithm that we do not have as a first
-     * option */
-    torture_update_sshd_config(state, sshd_config);
-
-    /* Make the client send the first_kex_packet_follows flag during key
-     * exchange as well as during the rekey */
-    s->ssh.session->send_first_kex_follows = true;
-
-    rc = ssh_options_set(s->ssh.session, SSH_OPTIONS_REKEY_DATA, &bytes);
-    assert_ssh_return_code(s->ssh.session, rc);
-
-    session_setup_sftp(state);
-
-    torture_rekey_recv(state);
-}
-#endif /* WITH_SFTP */
-
 int torture_run_tests(void) {
     int rc;
     struct CMUnitTest tests[] = {
@@ -780,34 +677,19 @@ int torture_run_tests(void) {
         cmocka_unit_test_setup_teardown(torture_rekey_different_kex,
                                         session_setup,
                                         session_teardown),
-        /* TODO verify the two rekey are possible and the states are not broken after rekey */
-
-        cmocka_unit_test_setup_teardown(torture_rekey_server_different_kex,
-                                        session_setup,
-                                        session_teardown),
-        /* Note, that these tests modify the sshd_config so follow-up tests
-         * might get unexpected behavior if they do not update the server with
-         * torture_update_sshd_config() too */
+        /* Note, that this modifies the sshd_config */
         cmocka_unit_test_setup_teardown(torture_rekey_server_send,
                                         session_setup,
                                         session_teardown),
-        cmocka_unit_test_setup_teardown(torture_rekey_guess_send,
-                                        session_setup,
-                                        session_teardown),
-        cmocka_unit_test_setup_teardown(torture_rekey_guess_wrong_send,
-                                        session_setup,
-                                        session_teardown),
 #ifdef WITH_SFTP
         cmocka_unit_test_setup_teardown(torture_rekey_server_recv,
                                         session_setup_sftp_server,
                                         session_teardown),
-        cmocka_unit_test_setup_teardown(torture_rekey_guess_recv,
-                                        session_setup,
-                                        session_teardown),
-        cmocka_unit_test_setup_teardown(torture_rekey_guess_wrong_recv,
-                                        session_setup,
-                                        session_teardown),
 #endif /* WITH_SFTP */
+        cmocka_unit_test_setup_teardown(torture_rekey_server_different_kex,
+                                        session_setup,
+                                        session_teardown),
+        /* TODO verify the two rekey are possible and the states are not broken after rekey */
     };
 
     ssh_init();
