@@ -4,10 +4,10 @@
 
 #define LIBSSH_STATIC
 #include <libssh/priv.h>
-#include "torture.h"
 
 #include "knownhosts.c"
 
+#include "torture.h"
 #if (defined _WIN32) || (defined _WIN64)
 #ifndef S_IRWXO
 #define S_IRWXO 0
@@ -574,13 +574,9 @@ static void torture_knownhosts_algorithms(void **state)
     char *algo_list = NULL;
     ssh_session session;
     bool process_config = false;
-    const char *expect = "ssh-ed25519,rsa-sha2-512,rsa-sha2-256,ssh-rsa,"
+    const char *expect = "ssh-ed25519,rsa-sha2-512,rsa-sha2-256,"
                          "ecdsa-sha2-nistp521,ecdsa-sha2-nistp384,"
-                         "ecdsa-sha2-nistp256"
-#ifdef HAVE_DSA
-                         ",ssh-dss"
-#endif
-    ;
+                         "ecdsa-sha2-nistp256";
     const char *expect_fips = "rsa-sha2-512,rsa-sha2-256,ecdsa-sha2-nistp521,"
                               "ecdsa-sha2-nistp384,ecdsa-sha2-nistp256";
 
@@ -613,13 +609,9 @@ static void torture_knownhosts_algorithms_global(void **state)
     char *algo_list = NULL;
     ssh_session session;
     bool process_config = false;
-    const char *expect = "ssh-ed25519,rsa-sha2-512,rsa-sha2-256,ssh-rsa,"
+    const char *expect = "ssh-ed25519,rsa-sha2-512,rsa-sha2-256,"
                          "ecdsa-sha2-nistp521,ecdsa-sha2-nistp384,"
-                         "ecdsa-sha2-nistp256"
-#ifdef HAVE_DSA
-                         ",ssh-dss"
-#endif
-    ;
+                         "ecdsa-sha2-nistp256";
     const char *expect_fips = "rsa-sha2-512,rsa-sha2-256,ecdsa-sha2-nistp521,"
                               "ecdsa-sha2-nistp384,ecdsa-sha2-nistp256";
 
